"""asterisk_mbox_server.py -- Do stuff."""

import sys
import socket
import select
import configparser
import logging
import json
import time
import argparse
import zlib

from threading import Thread
from asterisk_mbox_server.mbox import WatchMailBox
from asterisk_mbox_server.cdr import WatchCDR

import asterisk_mbox.commands as cmd
from asterisk_mbox.utils import (PollableQueue, recv_blocking,
                                 encode_password, compare_password,
                                 decode_from_sha)

__version__ = "0.5.2"


def _parse_request(msg):
    """Parse request from client."""
    request = {}
    request['cmd'] = msg[0]
    request['dest'] = msg[1]
    request['sha'] = msg[2:66].decode('utf-8')
    return request


class Connection(Thread):
    """Thread to handle a single TCP conection."""

    def __init__(self, conn, mbox, cdr, password):
        """Initialize."""
        # pylint: disable=too-many-arguments
        Thread.__init__(self)

        self.conn = conn
        self.mbox = mbox
        self.cdr = cdr
        self.mboxq = PollableQueue()
        self.password = password
        self.accept_pw = False

    def queue(self):
        """Fetch queue object."""
        return self.mboxq

    def _build_msg_list(self):
        """Send a sanitized list to the client."""
        result = []
        blacklist = ["wav"]
        status = self.mbox.get_mbox_status()
        for _k, ref in status.items():
            result.append({key: val for key, val in ref.items()
                           if key not in blacklist})
        return result

    @staticmethod
    def _build_cdr(entries, start=0, count=-1, sha=None):
        """Extract requested info from cdr list."""
        if sha is not None:
            msg = decode_from_sha(sha)
            start, count = [int(item) for item in msg.split(b',')]
        if not start or start < 0:
            start = 0
        if start >= len(entries) or count == 0:
            return msg

        if count < 0 or count + start > len(entries):
            end = len(entries)
        else:
            end = start + count
        return entries[start:end]

    def _send(self, command, msg):
        """Send message prefixed by length."""
        msglen = len(msg)
        msglen = bytes([msglen >> 24,
                        0xff & (msglen >> 16),
                        0xff & (msglen >> 8),
                        0xff & (msglen >> 0)])
        self.conn.send(bytes([command]) + msglen + msg)

    def _handle_request(self, request):
        if request['cmd'] == cmd.CMD_MESSAGE_PASSWORD:
            self.accept_pw, msg = compare_password(self.password,
                                                   request['sha'])
            if self.accept_pw:
                self._send(cmd.CMD_MESSAGE_VERSION,
                           __version__.encode('utf-8'))
                logging.info("Password accepted")
            else:
                self._send(cmd.CMD_MESSAGE_ERROR, msg.encode('utf-8'))
                logging.warning("Password rejected: %s", msg)
        elif not self.accept_pw:
            logging.warning("Bad Password")
            self._send(cmd.CMD_MESSAGE_ERROR, b'bad password')
        elif request['cmd'] == cmd.CMD_MESSAGE_LIST:
            logging.debug("Requested Message List")
            self._send(cmd.CMD_MESSAGE_LIST,
                       json.dumps(self._build_msg_list()).encode('utf-8'))
        elif request['cmd'] == cmd.CMD_MESSAGE_MP3:
            msg = self.mbox.mp3(request['sha'])
            if msg:
                self._send(cmd.CMD_MESSAGE_MP3, msg)
            else:
                logging.warning("Couldn't find message for %s", request['sha'])
                self._send(cmd.CMD_MESSAGE_ERROR,
                           "Could not find requested message")
        elif request['cmd'] == cmd.CMD_MESSAGE_CDR_AVAILABLE:
            if not self.cdr:
                self._send(cmd.CMD_MESSAGE_ERROR, b'CDR Not enabled')
                return
            self._send(cmd.CMD_MESSAGE_CDR_AVAILABLE,
                       json.dumps({'count': self.cdr.count()}).encode('utf-8'))
        elif request['cmd'] == cmd.CMD_MESSAGE_CDR:
            if not self.cdr:
                self._send(cmd.CMD_MESSAGE_ERROR, b'CDR Not enabled')
                return
            entries = self._build_cdr(self.cdr.entries(), sha=request['sha'])
            msg = {'keys': self.cdr.keys(), 'entries': entries}
            self._send(cmd.CMD_MESSAGE_CDR,
                       zlib.compress(json.dumps(msg).encode('utf-8')))

    def run(self):
        """Thread main loop."""
        while True:
            readable, _w, _e = select.select([self.conn, self.mboxq], [], [])
            if self.conn in readable:
                try:
                    request = _parse_request(recv_blocking(self.conn, 66))
                except RuntimeError:
                    logging.warning("Connection closed")
                    break
                logging.debug(request)
                self._handle_request(request)

            if self.mboxq in readable:
                msgtype = self.mboxq.get()
                self.mboxq.task_done()
                if msgtype == 'mbox':
                    self._send(cmd.CMD_MESSAGE_LIST,
                               json.dumps(self._build_msg_list())
                               .encode('utf-8'))
                elif msgtype == 'cdr':
                    self._send(cmd.CMD_MESSAGE_CDR_AVAILABLE,
                               json.dumps({'count': self.cdr.count()})
                               .encode('utf-8'))


class AsteriskMboxServer:  # pylint: disable=too-few-public-methods
    """Monitor Asterisk Mailbox."""

    def __init__(self):
        """Initialize."""
        args = self._parse_args()
        logging.basicConfig(level=logging.DEBUG
                            if args.verbose else logging.INFO)

        self._opts = self._config(args.cfg)
        self._soc = self._setup_socket()

        self._mailbox = WatchMailBox(self._opts['mbox_path'],
                                     self._opts['cache_file'],
                                     {'GOOGLE_KEY': self._opts['google_key']})
        self._mailbox.setDaemon(True)
        self._mailbox.start()

        self._cdr = None
        if self._opts['cdr']:
            self._cdr = WatchCDR(self._opts['cdr'],
                                 self._opts['poll_interval'])
            self._cdr.setDaemon(True)
            self._cdr.start()

        self._clients = []

    @staticmethod
    def _config(fname):
        config = configparser.ConfigParser()
        config.read(fname)
        config = config["default"]
        opts = {}
        password = config.get('password')
        opts['password'] = encode_password(password)
        opts['port'] = config.getint('port', 12345)
        opts['mbox_path'] = config.get('mbox_path')
        opts['cache_file'] = config.get('cache_file', 'asteriskmbox.pkl')
        opts['google_key'] = config.get('google_key')
        opts['ipaddr'] = config.get('ipaddr', '')
        opts['min_interval'] = config.getint('min_interval', 10)
        opts['cdr'] = config.get('cdr', None)
        opts['poll_interval'] = config.getint('poll_interval', 5)
        return opts

    @staticmethod
    def _parse_args():
        """Parse cmdline arguments."""
        parser = argparse.ArgumentParser()
        parser.add_argument("--cfg", help="Path to configuration file")
        parser.add_argument("-v", "--verbose", help="Turn on debug logging",
                            action='store_true')
        return parser.parse_args()

    def _setup_socket(self):
        soc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        soc.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            soc.bind((self._opts['ipaddr'], self._opts['port']))
        except socket.error:
            logging.exception('Bind failed.')
            sys.exit()
        soc.listen(10)
        return soc

    def _connection(self):
        """Initialize a new connection."""
        conn, addr = self._soc.accept()
        ipaddr, port = str(addr[0]), str(addr[1])
        logging.info('Accepting connection from ' + ipaddr + ':' + port)

        try:
            thread = Connection(conn, self._mailbox, self._cdr,
                                self._opts['password'])
            thread.setDaemon(True)
            thread.start()
            self._clients.append(thread)
        except Exception:                      # pylint: disable=broad-except
            logging.exception("Terible error!")

    def _handle_cdr_update(self):
        """Read CDR update."""
        queue = self._cdr.queue()
        queue.get()
        queue.task_done()
        for thread in self._clients:
            thread.queue().put('cdr')

    def _handle_mbox_update(self, timeout, last_time, readable):
        """Read Mailbox update."""
        read_mbox = False
        if readable:
            queue = self._mailbox.queue()
            queue.get()
            queue.task_done()
            if not timeout:
                if time.time() - last_time < self._opts['min_interval']:
                    last_time = time.time()
                    timeout = self._opts['min_interval']
                else:
                    read_mbox = True
        if timeout:
            if time.time() - last_time >= self._opts['min_interval']:
                read_mbox = True
            else:
                timeout = last_time + self._opts['min_interval'] - time.time()
        if read_mbox:
            self._mailbox.update()
            for thread in self._clients:
                thread.queue().put('mbox')
            last_time = time.time()
            timeout = None
        return timeout, last_time

    def loop(self):
        """Run main thread."""
        timeout = None
        last_time = 0

        sockets = [self._soc, self._mailbox.queue()]
        if self._cdr:
            sockets.append(self._cdr.queue())
        while True:
            readable, _w, _e = select.select(sockets, [], [], timeout)

            if self._soc in readable:
                self._connection()

            self._clients = [thread for thread in self._clients
                             if thread.isAlive()]

            if self._cdr and self._cdr.queue() in readable:
                self._handle_cdr_update()

            timeout, last_time = self._handle_mbox_update(
                timeout, last_time,
                self._mailbox.queue() in readable)

        self._soc.close()


def main():
    """Execute main application."""
    AsteriskMboxServer().loop()


if __name__ == '__main__':
    main()



"""Watch mailbox for new messages."""

import os
import pickle
import logging
import configparser
import hashlib
import glob
import subprocess
import tempfile

from threading import Thread, Lock

import speech_recognition as sr
import inotify.adapters

from asterisk_mbox.utils import PollableQueue


class WatchMailBox(Thread):
    """Thread to watch for new/removed messages in a mailbox."""

    def __init__(self, path, stt_file, sr_keys):
        """Initializie."""
        Thread.__init__(self)

        self.mbox_queue = PollableQueue()
        self.path = path
        self.stt_file = stt_file
        self.sr_keys = sr_keys
        self.subdirs = []
        self.speech = sr.Recognizer()
        self.status = {}
        self.cache = {}
        self.lock = Lock()

        if os.path.isfile(stt_file):
            with open(stt_file, 'rb') as infile:
                self.cache = pickle.load(infile)
        self._save_cache()
        self.inot = inotify.adapters.Inotify()
        for subdir in ('INBOX', 'Old', 'Urgent'):
            directory = os.path.join(self.path, subdir)
            if os.path.isdir(directory):
                self.subdirs.append(subdir)
                logging.debug("Watching Directory: %s", directory)
                self.inot.add_watch(directory.encode('utf-8'))
            else:
                logging.debug("Directory %s not found", directory)

    def _save_cache(self):
        """Save cache data."""
        with open(self.stt_file, 'wb') as outfile:
            pickle.dump(self.cache, outfile, pickle.HIGHEST_PROTOCOL)

    def _speech_to_text(self, fname):
        """Convert WAV to text."""
        txt = ""
        logging.debug('STT ' + fname)
        try:
            with sr.WavFile(fname) as source:
                audio = self.speech.record(source)   # read the entire WAV file
                txt += self.speech.recognize_google(
                    audio, key=self.sr_keys['GOOGLE_KEY'])
        except sr.UnknownValueError:
            txt += "Google Speech Recognition could not understand audio"
        except sr.RequestError as exception:
            txt += "Could not request results from " \
                   "Google Speech Recognition service; {0}".format(exception)
        logging.debug("Parsed %s --> %s", fname, txt)
        return txt

    def _sha_to_fname(self, sha):
        """Find fname from sha256."""
        for fname in self.status:
            if sha == self.status[fname]['sha']:
                return fname
        return None

    @staticmethod
    def _parse_msg_header(fname):
        """Parse asterisk voicemail metadata."""
        ini = configparser.ConfigParser()
        ini.read(fname)
        try:
            return dict(ini.items('message'))
        except configparser.NoSectionError:
            logging.exception("Couldn't parse: %s", fname)
            return

    @staticmethod
    def _sha256(fname):
        """Get SHA256 sum of a file contents."""
        sha = hashlib.sha256()
        with open(fname, "rb") as infile:
            for chunk in iter(lambda: infile.read(4096), b""):
                sha.update(chunk)
        return sha.hexdigest()

    def _build_mbox_status(self):
        """Parse all messages in a mailbox."""
        data = {}
        for subdir in self.subdirs:
            directory = os.path.join(self.path, subdir)
            logging.debug('Reading: ' + directory)
            if not os.path.isdir(directory):
                continue
            for filename in os.listdir(directory):
                filename = os.path.join(directory, filename)
                if not os.path.isfile(filename):
                    continue
                basename, ext = os.path.splitext(filename)
                logging.debug('Parsing: %s -- %s', basename, ext)
                if basename not in data:
                    data[basename] = {'mbox': subdir}
                if ext == '.txt':
                    data[basename]['info'] = self._parse_msg_header(filename)
                elif ext == '.wav':
                    data[basename]['sha'] = self._sha256(filename)
                    data[basename]['wav'] = filename
        for fname, ref in list(data.items()):
            if ('info' not in ref or 'sha' not in ref or
                    not os.path.isfile(ref['wav'])):
                logging.debug('Message is not complete: ' + fname)
                del data[fname]
                continue
            sha = ref['sha']
            if sha not in self.cache:
                self.cache[sha] = {}
            if 'txt' not in self.cache[sha]:
                self.cache[sha]['txt'] = self._speech_to_text(ref['wav'])
                self._save_cache()
            logging.debug("SHA (%s): %s", fname, sha)
            ref['text'] = self.cache[sha]['txt']
        self.status = data

    def delete(self, sha):
        """Delete message from mailbox."""
        # This does introduce a race condition between
        # Asterisk resequencing and us deleting, but there isn't
        # much we can do about it
        self.lock.acquire()
        self._build_mbox_status()
        fname = self._sha_to_fname(sha)
        for fil in glob.glob(fname + ".*"):
            os.unlink(fil)
        dirname = os.path.basename(fname)
        paths = {}
        for fil in glob.iglob(os.path.join(dirname + "msg[0-9]*")):
            base, = os.path.splitext(fil)
            paths[base] = None
        last_idx = 0
        rename = []
        with tempfile.TemporaryDirectory(dir=dirname) as tmp:
            for base in sorted(paths):
                if base != os.path.join(dirname + "msg%04d" % (last_idx)):
                    for fil in glob.glob(base + ".*"):
                        ext = os.path.splitext(fil)[1]
                        newfile = "msg%04d.%s" % (last_idx, ext)
                        logging.info("Renaming '" + fil + "' to '"
                                     + os.path.join(dirname, newfile) + "'")
                        rename.append(newfile)
                        os.link(fil, os.path.join(tempfile, newfile))
            for fil in rename:
                finalname = os.path.join(dirname, fil)
                if os.path.lexists(finalname):
                    os.unlink(finalname)
                os.rename(os.path.join(tmp, fil), finalname)
        self.lock.release()

    def mp3(self, sha):
        """Convert WAV to MP3 using LAME."""
        try:
            fname = self._sha_to_fname(sha)
            if not fname:
                return None
            logging.debug("Requested MP3 for %s ==> %s",
                          sha, fname)
            wav = self.status[fname]['wav']
            process = subprocess.Popen(["lame", "--abr", "24", "-mm",
                                        "-h", "-c",
                                        "--resample", "22.050",
                                        "--quiet", wav, "-"],
                                       stdout=subprocess.PIPE)
            result = process.communicate()[0]
            return result
        except OSError:
            logging.exception("Failed To execute lame")
            return

    def update(self):
        """Rebuild mailbox data."""
        self.lock.acquire()
        self._build_mbox_status()
        self.lock.release()

    def get_mbox_status(self):
        """Return current mbox status."""
        return self.status

    def queue(self):
        """Fetch queue object."""
        return self.mbox_queue

    def run(self):
        """Execute main loop."""
        self.mbox_queue.put("rebuild")
        trigger_events = ('IN_DELETE', 'IN_CLOSE_WRITE',
                          'IN_MOVED_FROM', 'IN_MOVED_TO')
        rebuild = False
        try:
            for event in self.inot.event_gen():
                if event is not None:
                    (_header, type_names, _watch_path, _filename) = event
                    if set(type_names) & set(trigger_events):
                        rebuild = True
                else:
                    if rebuild:
                        logging.debug("Rebuilding mailbox due to event: %s",
                                      type_names)
                        self.mbox_queue.put("rebuild")
                    rebuild = False
        finally:
            for subdir in self.subdirs:
                directory = os.path.join(self.path, subdir)
                self.inot.remove_watch(directory.encode('utf-8'))



"""Monitor cdr logs for changes."""
import os
import sys
import csv
import time
import logging
import posixpath
from threading import Thread
from asterisk_mbox.utils import PollableQueue


class WatchCDR(Thread):
    """Watch for new entries in the CDR database."""

    def __init__(self, cdr_path, sleep=5.0):
        """Initialize variables."""
        Thread.__init__(self)

        self._cdr_queue = PollableQueue()
        self._sleep = sleep
        self._inot = None
        self._watch = None
        self._sql = None
        self._cdr_file = None
        self._entries = []
        self._datefield = 'calldate'
        if not cdr_path:
            # Disable CDR handling
            self._keymap = {}
        elif os.path.isfile(cdr_path):
            self._cdr_file = cdr_path
            self._header = ['accountcode', 'src', 'dst', 'dcontext', 'clid',
                            'channel', 'dstchannel', 'lastapp', 'lastdata',
                            'start', 'answer', 'end', 'duration', 'billsec',
                            'disposition', 'amaflags', 'uniqueid', 'userfield']
            self._keymap = {
                'time': 'start',
                'callerid': 'clid',
                'src': 'src',
                'dest': 'lastdata',
                'context': 'dcontext',
                'application': 'lastapp',
                'duration': 'duration',
                'disposition': 'disposition',
                }
            try:
                if sys.platform.startswith('linux'):
                    import inotify.adapters
                    self._inot = inotify.adapters.Inotify()
                    self._inot.add_watch(self._cdr_file.encode('utf-8'))
            finally:
                pass
        else:
            import sqlalchemy
            try:
                self._sql = sqlalchemy.create_engine(
                    posixpath.dirname(cdr_path))
                self._query = sqlalchemy.text(
                    "SELECT * from {} ORDER BY `{}` DESC"
                    .format(posixpath.basename(cdr_path), self._datefield))
            except sqlalchemy.exc.OperationalError as exception:
                logging.error(exception)
            self._keymap = {
                'time': 'calldate',
                'callerid': 'clid',
                'src': 'src',
                'dest': 'lastdata',
                'context': 'dcontext',
                'application': 'lastapp',
                'duration': 'duration',
                'disposition': 'disposition',
                }
        self._keys = list(self._keymap.keys())

    def _update_entries(self, entries):
        entries = [{key: e.get(origkey, "")
                    for (key, origkey) in self._keymap.items()}
                   for e in entries]
        if (len(entries) == len(self._entries)
                and (not entries or entries[0] == self._entries[0])):
            return False
        self._entries = entries
        return True

    def _read_cdr_from_sql(self):
        result = self._sql.execute(self._query)
        # This isn't thread safe, but since the values should
        # virtually never change, is safe in reality.
        # keys = [str(key) for key in result.keys()]
        entries = [{key: str(value) for (key, value)
                    in row.items()} for row in result]
        return self._update_entries(entries)

    def _read_cdr_file(self):
        with open(self._cdr_file) as filp:
            lines = filp.readlines()
            keys = None
            entries = csv.reader(lines, quotechar='"', delimiter=',',
                                 quoting=csv.QUOTE_ALL,
                                 skipinitialspace=True)
            if entries:
                length = len(entries[0])
                if length < len(self._header):
                    keys = self._header[:length]
                else:
                    keys = self._header
                entries = [dict(zip(keys, e[:len(keys)])) for e in entries]
            return self._update_entries(reversed(entries))

    def _inotify_loop(self):
        updated = False
        trigger_events = ('IN_DELETE', 'IN_CLOSE_WRITE',
                          'IN_MOVED_FROM', 'IN_MOVED_TO')
        try:
            for event in self._inot.event_gen():
                if event is not None:
                    (_header, type_names, _watch_path, _filename) = event
                    if set(type_names) & set(trigger_events):
                        updated = True
                else:
                    if updated:
                        if self._read_cdr_file():
                            logging.debug(
                                "Reloading CDR file due to event: %s",
                                type_names)
                            self._cdr_queue.put("updated")
                    updated = False
        finally:
            self._inot.remove_watch(self._cdr_file.encode('utf-8'))

    def run(self):
        """Thread main loop."""
        if self._inot:
            self._inotify_loop()
            return
        last_mtime = 0
        while True:
            updated = False
            if self._sql:
                updated = self._read_cdr_from_sql()
            elif self._cdr_file:
                mtime = os.path.getmtime(self._cdr_file)
                if last_mtime != mtime:
                    updated = self._read_cdr_file()
                    last_mtime = mtime
            if updated:
                self._cdr_queue.put("updated")
            time.sleep(self._sleep)

    def entries(self):
        """Retrieve current CDR log."""
        return self._entries

    def keys(self):
        """Retrieve CDR entity keys."""
        return self._keys

    def count(self):
        """Retrieve number of current CDR entries."""
        return len(self._entries)

    def queue(self):
        """Fetch queue object."""
        return self._cdr_queue



from setuptools import setup, find_packages
import io
import os
import re


# Example code to pull version from esptool.py with regex, taken from
# https://packaging.python.org/guides/single-sourcing-package-version/
def read(*names, **kwargs):
    with io.open(
            os.path.join(os.path.dirname(__file__), *names),
            encoding=kwargs.get("encoding", "utf8")
    ) as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

setup(
  name = 'asterisk_mbox_server',
  packages=find_packages(),
  version=find_version("asterisk_mbox_server", "__init__.py"),
  description = 'The server portion of a client/server to interact with Asterisk voicemail mailboxes',
  long_description=open('README.rst').read(),
  author = 'PhracturedBlue',
  author_email = 'rc2012@pblue.org',
  url = 'https://github.com/PhracturedBlue/asterisk_mbox_server', # use the URL to the github repo
  keywords = ['testing', 'asterisk' 'mailbox', 'voicemail'], # arbitrary keywords
  classifiers = [],
  install_requires=[
        "inotify",
        "SpeechRecognition",
        "asterisk_mbox",
   ],
   entry_points = {
        'console_scripts': [
            'asterisk-mbox-server = asterisk_mbox_server:main'
        ]
   },

)
